import logging
from http import HTTPStatus

import rasax.community.utils.common as common_utils
from rasax.community.services.workspaces_service import (
    WorkspacesHasDefaultName,
    WorkspacesIdNotExists,
    WorkspacesNameExists,
    WorkspacesService,
)
from rasax.community.api.decorators import rasa_x_scoped, validate_schema
from sanic import Blueprint, response
from sanic.request import Request
from sanic.response import HTTPResponse

logger = logging.getLogger(__name__)


def blueprint() -> Blueprint:
    """Create endpoints related to Workspaces management.

    Returns:
        Sanic `Blueprint` with endpoints.
    """
    workspaces_endpoints = Blueprint("workspaces_endpoints")

    @workspaces_endpoints.route("/workspaces", methods=["GET", "HEAD"])
    @rasa_x_scoped("workspaces.list", allow_api_token=True)
    async def get_workspaces(request: Request) -> HTTPResponse:
        """Return a list of all existing workspaces.

        Args:
            request: Incoming HTTP request.

        Returns:
            HTTP response.
        """
        workspaces = WorkspacesService.from_request(request).get_workspaces()
        return response.json(workspaces)

    @workspaces_endpoints.route("/workspaces", methods=["POST"])
    @rasa_x_scoped("workspaces.create", allow_api_token=True)
    @validate_schema("workspace")
    async def create_workspace(request: Request) -> HTTPResponse:
        """Create a new workspace with a specified name.

        Args:
            request: Incoming HTTP request.

        Returns:
            HTTP response.
        """
        try:
            workspace = WorkspacesService.from_request(request).create_workspace(
                request.json["name"]
            )

            return response.json(workspace, HTTPStatus.CREATED)
        except WorkspacesNameExists as e:
            logger.error(e)
            return common_utils.error(
                HTTPStatus.CONFLICT,
                "WorkspaceCreationFailed",
                "Failed to create workspace.",
                details=e,
            )

    @workspaces_endpoints.route("/workspaces/<workspace_id>", methods=["PUT"])
    @rasa_x_scoped("workspaces.update", allow_api_token=True)
    @validate_schema("workspace")
    async def update_workspace(request: Request, workspace_id: int) -> HTTPResponse:
        """Updates the workspace name.

        Args:
            request: Incoming HTTP request.

        Returns:
            HTTP response.
        """
        try:
            workspace = WorkspacesService.from_request(request).update_workspace(
                workspace_id, request.json["name"]
            )

            return response.json(workspace, HTTPStatus.OK)
        except WorkspacesNameExists as e:
            logger.error(e)
            return common_utils.error(
                HTTPStatus.CONFLICT,
                "WorkspaceUpdateFailed",
                "Failed to update workspace name.",
                details=e,
            )
        except WorkspacesIdNotExists as e:
            logger.error(e)
            return common_utils.error(
                HTTPStatus.NOT_FOUND,
                "WorkspaceUpdateFailed",
                "Failed to update workspace name.",
                details=e,
            )
        except WorkspacesHasDefaultName as e:
            logger.error(e)
            return common_utils.error(
                HTTPStatus.BAD_REQUEST,
                "WorkspaceUpdateFailed",
                "Failed to update workspace name.",
                details=e,
            )

    return workspaces_endpoints
