import json
from typing import Any, Text, Dict, Union, List

import sqlalchemy as sa
from rasax.community.database import utils

from rasax.community.database.base import Base, Timestamps
from sqlalchemy.orm import relationship, validates

FLOW_ELEMENT_TYPES = [
    "step",
    "arrow",
]


class BusinessLogicFlow(Base, Timestamps):
    """Represents a Business Logic Flow of the assistant."""

    __tablename__ = "business_logic_flow"

    id = sa.Column(sa.Integer, utils.create_sequence(__tablename__), primary_key=True)
    name = sa.Column(sa.String, nullable=False, unique=True)
    elements = relationship("BusinessLogicFlowElement", cascade="all, delete-orphan")

    @validates("name")
    def validate_not_empty(self, key: Text, value: Any) -> Any:
        """Validate that columns have a value.

        If the column contains string values, ensure that it's not an empty string.
        """
        if value is None or value == "":
            raise ValueError(f"empty {key} is invalid")
        return value

    def as_dict(self) -> Dict[Text, Union[Text, List[Dict[Text, Any]]]]:
        """Converts the class into a dict."""
        return {
            "id": self.id,
            "name": self.name,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
            "elements": [json.loads(e.data) for e in self.elements],
        }


class BusinessLogicFlowElement(Base, Timestamps):
    """Represents a Business Logic Flow Element of the assistant."""

    __tablename__ = "business_logic_flow_element"

    id = sa.Column(sa.Integer, utils.create_sequence(__tablename__), primary_key=True)
    type = sa.Column(sa.String, nullable=False)
    element_id = sa.Column(sa.String, nullable=False)
    data = sa.Column(sa.Text, nullable=False)
    business_logic_flow_id = sa.Column(
        sa.Integer, sa.ForeignKey("business_logic_flow.id")
    )
    flow = relationship("BusinessLogicFlow")

    @validates("element_id")
    def validate_not_empty_element_id(self, _: Text, value: Text) -> Text:
        """Validate that element_id is not missing or an empty string."""
        if value is None or value == "":
            raise ValueError("empty element id is invalid")
        return value

    @validates("type")
    def validate_type(self, _: Text, value: Text) -> Text:
        """Validate that the type only consists of valid values."""
        if value not in FLOW_ELEMENT_TYPES:
            raise ValueError("invalid element_type value")
        return value


class ExampleConversation(Base, Timestamps):
    """Represents an Example Conversation of the assistant."""

    __tablename__ = "example_conversation"

    id = sa.Column(sa.Integer, utils.create_sequence(__tablename__), primary_key=True)
    name = sa.Column(sa.String, nullable=False, unique=True)
    items = relationship("ExampleConversationItem", cascade="all, delete-orphan")


class ExampleConversationItem(Base, Timestamps):
    """Represents an Example Conversation Item of the assistant."""

    __tablename__ = "example_conversation_item"

    id = sa.Column(sa.Integer, utils.create_sequence(__tablename__), primary_key=True)
    type = sa.Column(sa.String, nullable=False)
    text = sa.Column(sa.Text, nullable=False)
    example_conversation_id = sa.Column(
        sa.Integer, sa.ForeignKey("example_conversation.id")
    )
    conversation = relationship("ExampleConversation")
