import json
from typing import Text, List, Dict, Any

from rasax.community.database.assistant import (
    BusinessLogicFlow,
    BusinessLogicFlowElement,
)
from rasax.community.database.service import DbService
from sanic.request import Request
from sqlalchemy.exc import IntegrityError


class AssistantService(DbService):
    """A service used to work with assistant-related models."""

    @staticmethod
    def from_request(request: Request, **kwargs) -> "AssistantService":
        """Construct `InsightService` from an incoming HTTP request.

        Args:
            request: Incoming HTTP request.
            **kwargs: Other key-value args, not used.

        Returns:
            Constructed `InsightService` object.
        """
        return AssistantService(request.ctx.db_session)

    def create_flow(
        self, name: Text, elements: List[Dict[Text, Any]]
    ) -> BusinessLogicFlow:
        """Creates a new Business Logic Flow.

        Args:
            name: A unique name.
            elements: a list of flowchart steps and arrows as dicts.

        Raises:
            BusinessLogicFlowParseError: When parsing of the elements fail.

        Returns:
            Created `BusinessLogicFlow`.
        """
        try:
            flow = BusinessLogicFlow(name=name)
            flow.elements = self._parse_flow_elements(elements)
            self.add(flow)

            self.flush()
            return flow
        except IntegrityError:
            self.rollback()
            raise DuplicateBusinessLogicFlowException()

    @staticmethod
    def _parse_flow_elements(
        elements: List[Dict[Text, Any]]
    ) -> List[BusinessLogicFlowElement]:
        """Parses flow elements and returns initialized `BusinessLogicFlowElement` objects."""
        flow_elements = []

        for element in elements:
            type_ = element.get("data").get("element_type")
            element_id = element.get("id")
            data = json.dumps(element)

            flow_elements.append(
                BusinessLogicFlowElement(type=type_, element_id=element_id, data=data)
            )

        return flow_elements


class DuplicateBusinessLogicFlowException(Exception):
    """Exception for a business logic flow with a duplicate name."""

    pass
